<?php
/**
 * Plugin Name: Exodox
 * Plugin URI:  https://exodox.link/get-plugin/
 * Version:     0.8.3
 * Author:      Exodox
 * Author URI:  https://exodox.link
 * License:     Exodox
 * License URI: https://exodox.link/license
 * Text Domain: exodox
 * Domain Path: /languages
 * Description: Exodox content lock plugin.
 *
 * @package exodox
 */

/*
 * Helper constants.
 */
if (!defined('EXODOX_APP_HOST')) {
    define('EXODOX_APP_HOST', 'app.exodox.link');
}
if (!defined('EXODOX_API_HOST')) {
    define('EXODOX_API_HOST', 'api.exodox.link');
}
if (!defined('EXODOX_DEV')) {
    define('EXODOX_DEV', false);
}
if (!defined('EXODOX_PROTO')) {
    define('EXODOX_PROTO', 'https://');
}

define('EXODOX_QUERY_DEBUG', false);
define('EXODOX_QUERY_DEBUG_STYLE', 'clear: both; color: #fff; font-weight: bold; background: #333;');
define('EXODOX_QUERY_STATUS_SINGLE', 'one_matching_query_found');
define('EXODOX_QUERY_STATUS_MULTIPLE', 'multiple_matching_query_found');
define('EXODOX_QUERY_STATUS_NONE', 'no_matching_query_found');

/**
 * Class ExodoxLogin, break out into more classes. Something like "ExodoxAuth", "ExodoxUrl" and "ExodoxPost".
 */
class ExodoxLogin
{
    const EXODOX_PLUGIN_VERSION = '0.8.3';
    const EXODOX_ADMIN_PAGE_SLUG = 'exodox';
    const EXODOX_COOKIE_NAME = 'exodoxUserId';
    const EXODOX_HOST_URL = EXODOX_APP_HOST;
    const EXODOX_LOGIN_PAGE = EXODOX_PROTO.EXODOX_APP_HOST.'/signup/consumer';
    const EXODOX_LOGIN_CHECK_PAGE = EXODOX_PROTO.EXODOX_APP_HOST.'/api/check-login';
    const EXODOX_CHECK_LOGIN_RETURN_PAGE = 'exodoxCheckLoginReturnPage';
    const EXODOX_DOUBLE_CHECK_LOGIN_RETURN_PAGE = 'exodoxRedirectReturnPage';
    const EXODOX_TEASER_PAGE = EXODOX_PROTO.EXODOX_APP_HOST.'/';
    const EXODOX_CHECK_LOCKED_POSTS_ENDPOINT = EXODOX_PROTO.EXODOX_API_HOST.'/v1/api/links';
    const EXODOX_CHECK_LOCKED_POSTS_WITH_UUID_ENDPOINT = EXODOX_PROTO.EXODOX_API_HOST.'/v1/api/readerLinks';
    const EXODOX_API_TIMEOUT = 7; // Seconds
    const EXODOX_URL_EXCEPTIONS = ['xmlrpc.php']; // Pages that should never trigger any api calls
    // TODO: Research and implement reasonable timeout for user cookie
    // const EXODOX_USER_COOKIE_EXPIRE = 3600; // Seconds

    private static $exodoxUserID = '';
    private static $exodoxCurrentPosts = [];
    private static $exodoxLockedPosts = [];
    private static $exodoxLockedPostsObjects = [];
    private static $exodoxSingularTitlesReplaced = [];
    private static $exodoxQueries = []; // Memory of queries used by WordPress' "the loop"
    private static $exodoxReadMoreKeys = []; // Memory of read more elements shown; to prevent duplicates

    /**
     * Register WP actions and filters.
     */
    public static function initialize()
    {
        add_action('admin_init', [__CLASS__, 'admin_init']);
        add_action('admin_menu', [__CLASS__, 'admin_menu'], 99);
        add_filter('plugin_action_links_'.plugin_basename(__FILE__), [__CLASS__, 'plugin_action_links']);
        add_filter('exodox_publisher_id', [__CLASS__, 'filter_publisher_id']);
        add_action('plugins_loaded', [__CLASS__, 'exodox_load_plugin_textdomain']);
        add_filter('rest_authentication_errors', [__CLASS__, 'rest_authentication_errors']);
        add_action('wp_logout', [__CLASS__, 'exodox_clear_cookies']);
        add_action('rest_api_init', [__CLASS__, 'exodox_rest_api_init']);

        /* If we have a publisher id (and this is not a crawler) add non-admin actions and filters */
        if (self::get_publisher_id() && !self::is_bot() && !self::is_admin_page()) {
            add_action('init', [__CLASS__, 'init']);

            // Link filters
            add_filter('home_url', [__CLASS__, 'add_cookies_not_allowed_query_arg'], 10, 3);
            add_filter('post_link', [__CLASS__, 'add_cookies_not_allowed_query_arg'], 10, 3);
            add_filter('page_link', [__CLASS__, 'add_cookies_not_allowed_query_arg'], 10, 3);
            add_filter('attachment_link', [__CLASS__, 'add_cookies_not_allowed_query_arg'], 10, 3);
            add_filter('post_type_link', [__CLASS__, 'add_cookies_not_allowed_query_arg'], 10, 3);

            // Redirect filter
            add_filter('allowed_redirect_hosts', [__CLASS__, 'add_exodox_host']);

            // Post filters
            add_filter('the_posts', [__CLASS__, 'the_posts']);
            add_filter('the_title', [__CLASS__, 'replace_title_if_locked']);

            add_filter('loop_start', [__CLASS__, 'loop_start'], 0, 2);
            add_filter('the_post', [__CLASS__, 'add_read_more_element'], 0, 2);

            add_filter('the_content_feed', [__CLASS__, 'hide_content_in_feed'], 0, 2);
        }
    }

    /*********************************************/
    /* Wordpress actions (must be public static) */
    /*********************************************/

    /**
     * Register plugin settings form.
     */
    public static function admin_init()
    {
        if (!is_ssl()) {
            self::log_error('Wordpress admin accessed without SSL enabled. Disabling Exodox plugin.');
            add_action('admin_notices', [__CLASS__, 'admin_ssl_notice']);
            deactivate_plugins(plugin_basename(__FILE__));
            if (array_key_exists('activate', $_GET)) {
                unset($_GET['activate']);
            }
        }
        $settings = [
            'exodox_publisher_id' => [__CLASS__, 'filter_publisher_id'],
            'exodox_disable_listing' => 'sanitize_key',
        ];

        foreach ($settings as $option_name => $sanitize_callback) {
            register_setting(self::EXODOX_ADMIN_PAGE_SLUG, $option_name, $sanitize_callback);
        }

        add_settings_section(
            'settings',
            esc_html__('Settings', 'exodox'),
            __CLASS__.'::admin_settings_descr_callback',
            self::EXODOX_ADMIN_PAGE_SLUG
        );

        add_settings_field(
            'exodox_publisher_id',
            esc_html__('Publisher ID', 'exodox'),
            [__CLASS__, 'input_field'],
            self::EXODOX_ADMIN_PAGE_SLUG,
            'settings',
            [
                'name' => 'exodox_publisher_id',
                'type' => 'text',
                'class' => 'regular-text',
                'placeholder' => '',
            ]
        );

        add_settings_field(
            'exodox_disable_listing',
            esc_html__('Disable Exodox graphics on listing pages', 'exodox'),
            [__CLASS__, 'exodox_disable_listing_checkbox'],
            self::EXODOX_ADMIN_PAGE_SLUG,
            'settings'
        );
    }

    /**
     * Add admin menu page.
     */
    public static function admin_menu()
    {
        add_submenu_page(
            'options-general.php',
            esc_html__('Exodox - Lock content', 'exodox'),
            esc_html__('Exodox', 'exodox'),
            'manage_options',
            self::EXODOX_ADMIN_PAGE_SLUG,
            [__CLASS__, 'render_page']
        );
    }

    /**
     * Init, intercept early to handle returning users and redirects.
     */
    public static function init()
    {
        if (array_key_exists(self::EXODOX_COOKIE_NAME, $_COOKIE)) {
            self::$exodoxUserID = $_COOKIE[self::EXODOX_COOKIE_NAME];
        }

        self::request_handler();

        wp_register_style('exodoxCss', plugin_dir_url(dirname(__FILE__)).'exodox/css/screen.css', [], self::EXODOX_PLUGIN_VERSION);
        wp_enqueue_style('exodoxCss');
    }

    /**
     * Render error message about https not being enabled.
     */
    public static function admin_ssl_notice()
    {
        $notice = sprintf(
            'Exodox plugin requires SSL to work. If you feel unsure about how to enable SSL, you can use a plugin such as %1$sReally Simple SSL%2$s.',
            '<a href="'.esc_url('https://really-simple-ssl.com/').'">',
            '</a>'
        );
        $tip = sprintf(
            'If you look for a set of helpful instructions, %1$sthis article%2$s on wpbeginner.com may be what you are looking for.',
            '<a href="'.esc_url('https://www.wpbeginner.com/wp-tutorials/how-to-add-ssl-and-https-in-wordpress/').'">',
            '</a>'
        );

        echo <<<HTML
            <div class="error">
                <p>$notice</p>
                <p>$tip</p>
            </div>
HTML;
    }

    /*********************************************/
    /* Wordpress filters (must be public static) */
    /*********************************************/

    /**
     * Add settings link to plugin on plugin list in the WP admin.
     *
     * @param array $links existing plugin action links
     *
     * @return array
     */
    public static function plugin_action_links($links)
    {
        $settings_link = sprintf(
            '<a href="%s">%s</a>',
            esc_url(admin_url('options-general.php?page='.self::EXODOX_ADMIN_PAGE_SLUG)),
            esc_html__('Settings', 'exodox')
        );

        $links [] = $settings_link;

        return $links;
    }

    /**
     * Filter hook for publisher UUID input.
     *
     * @param $id
     *
     * @return string
     */
    public static function filter_publisher_id($id)
    {
        if (!wp_is_uuid($id)) {
            add_settings_error(
                'exodox_publisher_id',
                'exodox_publisher_id_error',
                esc_html__('Wrong format of Publisher Id', 'exodox'),
                'error'
            );
        }

        if (count(get_settings_errors('exodox_publisher_id'))) {
            return self::get_publisher_id();
        }

        return apply_filters('filter_publisher_id', sanitize_text_field($id));
    }

    /**
     * Loads text domain.
     *
     * @return null
     */
    public static function exodox_load_plugin_textdomain()
    {
        load_plugin_textdomain('exodox', false, basename(dirname(__FILE__)).'/languages/');
    }

    /**
     * Disable access to the public rest API.
     *
     * @param $result WP_Error|true|null
     *
     * @return WP_Error|true|null
     */
    public static function rest_authentication_errors($result)
    {
        if (true === $result || is_wp_error($result) || current_user_can('edit_posts') || preg_match('/\/wp-json\/exodox\/v1\/verify/i', $_SERVER['REQUEST_URI'])) {
            return $result;
        }

        $whitelist = ['127.0.0.1'];
        $remote_addr = $_SERVER ['REMOTE_ADDR'];

        if (!in_array($remote_addr, $whitelist)) {
            return new WP_Error(
                'forbidden_access',
                __('Access denied.'),
                [
                    'status' => 403,
                ]
            );
        }

        return $result;
    }

    /**
     * Clear Exodox cookies.
     *
     * @return null
     */
    public static function exodox_clear_cookies()
    {
        setcookie(self::EXODOX_COOKIE_NAME, '', time() - (60 * 60 * 24), '/');
    }

    /**
     * POST endpoint returning status 200 if the token matches the publisher id; otherwise 403.
     *
     * @param $data array Either an array of GET params or POST data
     *
     * @return WP_REST_Response|WP_Error
     */
    public static function exodox_verify_owner($data)
    {
        if ($data['token'] === md5(self::get_publisher_id())) {
            return new WP_REST_Response([
                'status' => 200,
                'response' => 'OK',
            ]);
        }

        return new WP_Error('not_verified', __('Could not verify ownership.'), [
            'status' => 401,
        ]);
    }

    /**
     * Implementation of the rest_api_init hook.
     *
     * @return null
     */
    public static function exodox_rest_api_init()
    {
        register_rest_route('exodox/v1', '/verify', [
            'methods' => 'POST',
            'callback' => [__CLASS__, 'exodox_verify_owner'],
            'permission_callback' => '__return_true',
        ]);
    }

    /**
     * Adds query argument to link if cookies are not allowed.
     *
     * @param $permalink
     *
     * @return string
     */
    public static function add_cookies_not_allowed_query_arg($permalink)
    {
        $cookieParameters = [];

        if (!self::cookies_allowed()) {
            $cookieParameters ['cookiesNotAllowed'] = 1;
        }

        return esc_url(add_query_arg($cookieParameters, $permalink));
    }

    /**
     * Register Exodox URL as an allowed redirect host. Used by wp_safe_redirect().
     *
     * @param array $hosts
     *
     * @return array $hosts
     */
    public static function add_exodox_host($hosts)
    {
        $exodoxHosts = [self::EXODOX_HOST_URL];

        return array_merge($hosts, $exodoxHosts);
    }

    /**
     * Filter that checks for and filters locked posts at load time.
     *
     * @param $posts
     *
     * @return array (filtered posts)
     */
    public static function the_posts($posts)
    {
        if (0 === count($posts)) {
            return $posts;
        }

        $newPosts = [];
        $returnPosts = $posts;

        foreach ($posts as $post) {
            $link = self::get_common_url(get_permalink($post->ID));
            if (!array_search($link, self::$exodoxCurrentPosts) && !array_search($link, $newPosts)) {
                $newPosts [] = $link;
            }
        }

        if (count($newPosts)) {
            $lockedPosts = self::retrieve_locked_posts($newPosts);

            foreach ($lockedPosts ['links'] as $lockedPost) {
                $queryLink = $lockedPost ['queryLink'];
                if (!in_array($queryLink, self::$exodoxLockedPosts, true)) {
                    self::$exodoxLockedPostsObjects[$queryLink] = $lockedPost;
                    array_push(self::$exodoxLockedPosts, $queryLink);
                }
            }
        }

        if (is_singular() && 0 === count(self::$exodoxCurrentPosts)) {
            $link = self::get_common_url(get_permalink($posts[0]->ID));

            if (self::is_path_locked($link)) {
                // Locked single post
                $lockedPost = self::lock_single_post($posts[0]);
                if ($lockedPost->is_exodox_locked) {
                    self::enqueue_lock_single_css($lockedPost);
                }
                $returnPosts = [$lockedPost];
            }
        }

        // Listing
        if (self::wrap_listing() && !is_singular() && count(self::$exodoxLockedPosts)) {
            foreach ($returnPosts as &$post) {
                $link = self::get_common_url(get_permalink($post->ID));
                if (self::is_path_locked($link) && !in_array($link, self::$exodoxCurrentPosts)) {
                    $lockStatus = self::get_lock_status($link);

                    if ('unlocked' !== $lockStatus) {
                        $lockedPostOjbect = self::get_locked_post_object($link);
                        $post->post_content = $lockedPostOjbect ['preamble'];
                        $post->post_excerpt = $lockedPostOjbect ['preamble'];
                    }
                }
            }
        }

        // Mark new posts as checked
        self::$exodoxCurrentPosts = array_merge(self::$exodoxCurrentPosts, $newPosts);

        return $returnPosts;
    }

    /**
     * Hides locked content title in some views, title is re-added together with the content later.
     *
     * @param $title (string)
     *
     * @return string
     */
    public static function replace_title_if_locked($title)
    {
        if (self::is_path_locked()) {
            $lockStatus = self::get_lock_status(get_permalink());
            switch ($lockStatus) {
                case 'unlocked':
                case 'creator':
                    return $title;
                case 'not-locked':
                    $lockedPostObject = self::get_locked_post_object(get_permalink());
                    self::log_warning("ERROR: Malformed locked link object\n".self::format_data_raw($lockedPostObject));

                    return $title;
                default:
                    $lockedPostObject = self::get_locked_post_object(get_permalink());
                    if (!$lockedPostObject) {
                        self::log_warning('ERROR: Path is locked but has no data on being locked!');

                        return $title;
                    }
                    if (is_singular() && !self::singular_title_replaced($lockedPostObject ['slug'])) {
                        self::replace_singular_title($lockedPostObject ['slug']); // Make sure the title is only replaced once
                        return "<div class='exodox-font-size-zero'>$title</div>";
                    }

                    return $title;
            }
        }

        return $title;
    }

    /**
     * The following sets of methods are added because of the unpredictability of theme creator's use
     * of WordPress' "the loop" (and its different supporting actions).
     *
     * In our case; we are adding a filter on the "the_post" method to inject the "read more" elements.
     * The issue with this is that this method is not exclusively used for displaying the post itself.
     * Some theme creator's use it to, i.e.:
     *      - display meta data about the post
     *      - show related post
     *      - suggest other posts of interest
     *      etc.
     * We do not want to add the "read more" element for all of these cases (although some qualify) and
     * in order to create some sort of long-term logic to this we need to know in what context a post is
     * shown and if we have already displayed an element for it.
     */

    /**
     * Adds the queries used by "the loop" to memory in order to match a post to a specific query.
     *
     * @since 0.7.1
     *
     * @param WP_Query $query
     *
     * @return null
     */
    public static function loop_start($query)
    {
        if (self::wrap_listing()) {
            self::$exodoxQueries [] = $query;

            // Debug of the loop in order to verify themes
            if (EXODOX_QUERY_DEBUG) {
                echo ''.
                '<pre style="'.EXODOX_QUERY_DEBUG_STYLE.'">'."\n".
                    'BEGIN LOOP'."\n".
                    'id: '.md5($query->request)."\n".
                    'posts: '.$query->post_count."\n".
                '</pre>'."\n";
            }
        }
    }

    /**
     * Gets statistics about the active query used from within the loop (i.e. from a "the_post" filter).
     *
     * @since 0.7.1
     *
     * @param WP_Query $query
     *
     * @return object
     */
    public static function get_loop_stats($query)
    {
        $matching_queries = array_filter(self::$exodoxQueries, function ($array_query) use ($query) {
            return md5($array_query->request) === md5($query->request); // Not a definite comparison due to the lack of a unique identifier in loops
        });

        if (1 === count($matching_queries)) {
            $matching_key = key($matching_queries);
            $query_keys = array_keys(self::$exodoxQueries);
            $matching_depth = array_search($matching_key, $query_keys);
            $matching_query = reset($matching_queries);
            $matching_id = md5($matching_query->request);
            $matching_in_the_loop = $matching_query->in_the_loop;

            return (object) [
                'status' => EXODOX_QUERY_STATUS_SINGLE,
                'id' => $matching_id,
                'key' => $matching_key,
                'depth' => $matching_depth,
                'in_the_loop' => $matching_in_the_loop,
                'total_queries' => count($query_keys),
            ];
        } elseif (count($matching_queries)) {
            return (object) [
                'status' => EXODOX_QUERY_STATUS_MULTIPLE,
            ];
        }

        return (object) [
            'status' => EXODOX_QUERY_STATUS_NONE,
            'error' => 'No matching queries found',
        ];
    }

    /**
     * Utility to decide if the "read more" element should be shown; or not.
     *
     * This is where most "theme hacks" should be placed.
     *
     * @since 0.7.1
     *
     * @param WP_Query $query
     *
     * @return bool
     */
    public static function should_display_read_more($query)
    {
        $link = self::get_common_url(get_permalink());
        $stats = self::get_loop_stats($query);

        if (EXODOX_QUERY_STATUS_SINGLE === $stats->status) {
            $unique_key = $stats->key.'-'.$stats->depth.'-'.md5($query->request).'-'.md5($link);

            // Heavy duty debug of the loop in order to verify themes
            if (EXODOX_QUERY_DEBUG) {
                echo ''.
                '<pre style="'.EXODOX_QUERY_DEBUG_STYLE.'">'."\n".
                    'stats: '.$stats->status.' '.$stats->key.' '.$stats->depth.' '.$stats->total_queries.' '.$stats->id.' '.$stats->in_the_loop."\n".
                    'post: '.md5($query->request).' '.$query->current_post.' '.$link."\n".
                    'key: '.$unique_key."\n".
                    'action: '.(in_array($unique_key, self::$exodoxReadMoreKeys) ? 'hide' : 'show')."\n".
                '</pre>'."\n";
            }

            if (in_array($unique_key, self::$exodoxReadMoreKeys)) { // Show only once per post & query
                return false;
            }

            self::$exodoxReadMoreKeys [] = $unique_key;

            return true;
        }

        return false;
    }

    /**
     * Adds "read more" element for locked posts in listings.
     *
     * @param array|string $content
     * @param WP_Query     $query
     *
     * @return array|string
     */
    public static function add_read_more_element($content, $query)
    {
        if (self::wrap_listing() && self::is_path_locked() && !is_singular() && self::should_display_read_more($query) && !is_feed()) {
            $link = self::get_common_url(get_permalink());
            $status = self::get_lock_status($link);
            $lockedPostObject = self::get_locked_post_object($link);

            $price = (int) $lockedPostObject ['price'] * 0.01;
            $currency = $lockedPostObject ['currency'];
            $text = sprintf(esc_html__('Read the whole article for only %1$s %2$s', 'exodox'), $price, $currency);

            switch ($status) {
                case 'not-locked': // Should not get here; but if we do, we exit
                    return $content;
                case 'creator':
                    $text = esc_html__('The article is locked with Exodox; but you, as administrator, can still read it.', 'exodox');
                    break;
                case 'unlocked':
                    $validTime = self::get_time_diff($lockedPostObject ['expiresAt']);
                    $text = sprintf(esc_html__('The article is unlocked, available for %1$s more', 'exodox'), $validTime);
                    break;
                case 'no-cookies':
                    $text = esc_html__('Locked content. You must allow cookies in order to be able to purchase this article.', 'exodox');
                    break;
                default: // Expected to catch "no-user", "locked" & "expired"
            }

            $icon = plugins_url('/images/seal.svg', __FILE__);

            $container = <<<HTML
                <div class="exodox_readmore">
                    <a href="$link" class="exodox_readmore_icon_link">
                        <img alt="" src="$icon" class="exodox_readmore_icon" />
                    </a>
                    <a href="$link" class="exodox_readmore_text" data-testid="$status">$text</a>
                </div>
HTML;

            echo $container;
        }

        return $content;
    }

    /**
     * Hide content in feeds of different types.
     *
     * @param string $content
     * @param string $feed_type
     *
     * @return string
     */
    public static function hide_content_in_feed($content, $type)
    {
        if (self::is_path_locked()) {
            return esc_html__('The article is locked with Exodox; and disabled in feeds.', 'exodox');
        }

        return $content;
    }

    /**********************************************/
    /* Wordpress sections (must be public static) */
    /**********************************************/

    /**
     * Render data about Exodox hosts when not in a production environment.
     */
    public static function admin_settings_descr_callback()
    {
        if (defined('EXODOX_APP_HOST') && EXODOX_APP_HOST !== 'app.exodox.link') {
            $output = <<<HTML
                <ul>
                    <li>EXODOX_APP_HOST: %s</li>
                    <li>EXODOX_API_HOST: %s</li>
                </ul>
HTML;

            echo sprintf($output, EXODOX_APP_HOST, EXODOX_API_HOST);
        }
    }

    /***********************************************/
    /* Helper functions (should be private static) */
    /***********************************************/

    /* Main route logic */

    /**
     * Get user, request and environment status and handle accordingly.
     */
    private static function request_handler()
    {
        // Step 1: Check if Exodox login has any relevance
        if (self::no_exodox_actions()) {
            return;
        }

        // Step 2: Check if we should redirect to https
        if (!is_ssl()) {
            self::redirect_to_https();
        }

        // Step 3: Check if we are at the login endpoint
        if (self::is_login_check_return_page()) {
            self::check_login_return_parameters();

            if (!self::cookies_allowed()) {
                self::redirect_to_target_url_no_cookies();

                return;
            }

            self::set_user_cookie();
            self::redirect_to_target_url();

            return;
        }

        // Step 4: Check if the user has been previously redirected
        if (!array_key_exists('cookiesNotAllowed', $_GET) && !array_key_exists(self::EXODOX_COOKIE_NAME, $_COOKIE)) {
            self::redirect_to_login_check();

            return;
        }
    }

    /* Simple state (non-mutating) checks */

    /**
     * Check if user is logged in.
     *
     * @return bool
     */
    public static function is_user_logged_in()
    {
        return '' !== self::get_user_id();
    }

    /**
     * Check if request is made by something that identifies as a bot, scraper etc.
     *
     * @uses $_SERVER ['HTTP_USER_AGENT']
     *
     * @return bool
     */
    public static function is_bot()
    {
        return
            array_key_exists('HTTP_USER_AGENT', $_SERVER)
            && preg_match('/exodox|Exodox Scrape-bot|facebook|axios|find|get|urllib|fetch|share|bot|crawl|slurp|spider|mediapartners/i', $_SERVER ['HTTP_USER_AGENT']);
    }

    /**
     * Checks if cookies are allowed by trying to access cookies previously set.
     *
     * @uses $_COOKIE ['exodoxUserId'], $_GET ['userId']
     *
     * @return bool
     */
    public static function cookies_allowed()
    {
        return array_key_exists(self::EXODOX_COOKIE_NAME, $_COOKIE) || array_key_exists('userId', $_GET);
    }

    /**
     * Check if request is for an admin-page.
     *
     * @return bool
     */
    public static function is_admin_page()
    {
        return (bool) is_admin();
    }

    /**
     * Check if request is from wordpress editor/admin.
     *
     * return bool
     */
    public static function user_can_edit_posts()
    {
        return is_user_logged_in() && current_user_can('edit_posts');
    }

    /**
     * Check if request is for a login-page.
     *
     * @return bool
     */
    public static function is_login_page()
    {
        $wp_login_regex = '|'.wp_login_url().'|';

        return preg_match($wp_login_regex, self::get_current_url());
    }

    /**
     * Check if request is for a page that should not trigger any actions.
     *
     * @return bool
     */
    private static function is_exempted_url()
    {
        foreach (self::EXODOX_URL_EXCEPTIONS as $url) {
            $url_regex = '|'.$url.'|';
            if (preg_match($url_regex, self::get_current_url())) {
                return true;
            }
        }

        return false;
    }

    /**
     * Check if the request is in an unsupported method.
     *
     * @uses $_SERVER ['REQUEST_METHOD']
     *
     * @return bool
     */
    private static function is_unsupported_method()
    {
        return 'get' !== strtolower($_SERVER ['REQUEST_METHOD']);
    }

    /**
     * Check if request is for login check return page.
     *
     * @return bool
     */
    public static function is_login_check_return_page()
    {
        $login_check_returnpage_regex = '|'.self::EXODOX_CHECK_LOGIN_RETURN_PAGE.'|';

        return preg_match($login_check_returnpage_regex, self::get_current_url());
    }

    /**
     * Check if current path is locked.
     *
     * @param $path string optional path argument
     *
     * @return bool
     */
    public static function is_path_locked($path = null)
    {
        if (null === $path) {
            $path = self::get_common_url(get_permalink());
        }

        if (in_array($path, self::$exodoxLockedPosts)) {
            $lockedPostObject = self::get_locked_post_object($path);

            if ($lockedPostObject && $lockedPostObject ['active']) {
                return true;
            }
        }

        return false;
    }

    /**
     * Returns true if the title has already been replaced with an inert version.
     *
     * @param $slug
     *
     * @return bool
     */
    protected static function singular_title_replaced($slug)
    {
        return array_key_exists($slug, self::$exodoxSingularTitlesReplaced);
    }

    /* Simple getters */

    /**
     * Check for user id in cookie.
     *
     * @return string
     */
    public static function get_user_id()
    {
        if (self::user_can_edit_posts() && self::get_publisher_id()) {
            return self::get_publisher_id();
        }

        if (36 === strlen(self::$exodoxUserID)) {
            return self::$exodoxUserID;
        }

        return '';
    }

    /**
     * Get the current URL.
     *
     * @uses $_SERVER ['HTTPS'], $_SERVER ['HTTP_HOST'], $_SERVER ['REQUEST_URI']
     *
     * @return bool
     */
    public static function get_current_url()
    {
        return (array_key_exists('HTTPS', $_SERVER) && 'on' === $_SERVER ['HTTPS'] ? 'https' : 'http')."://${_SERVER['HTTP_HOST']}${_SERVER['REQUEST_URI']}";
    }

    /**
     * Get the publisher id from admin input.
     *
     * @return string
     */
    public static function get_publisher_id()
    {
        return get_option('exodox_publisher_id');
    }

    /**
     * Should the plugin wrap posts on listing pages; or not.
     *
     * @since 0.7.8
     *
     * @return bool
     */
    public static function wrap_listing()
    {
        return 'yes' !== get_option('exodox_disable_listing');
    }

    /**
     * Check if this is an early exit case.
     *
     * @uses $_GET ['cookiesNotAllowed']
     */
    private static function no_exodox_actions()
    {
        return self::is_admin_page() ||
               self::is_login_page() ||
               self::is_exempted_url() ||
               self::is_unsupported_method();
    }

    /* Simple setters */

    /**
     * Sets user id cookie from GET parameter if present.
     *
     * @uses $_GET ['userId']
     */
    public static function set_user_cookie()
    {
        if (array_key_exists('userId', $_GET) && wp_is_uuid($_GET ['userId']) && self::get_user_id() != $_GET ['userId']) {
            $userId = sanitize_key($_GET ['userId']);

            setcookie('exodoxUserId', $userId, 0, '/');
            self::$exodoxUserID = $userId;
        }
    }

    /**
     * Signal to the plugin that the title has been replaced with an inert version.
     *
     * @param $slug
     */
    protected static function replace_singular_title($slug)
    {
        self::$exodoxSingularTitlesReplaced[$slug] = true;
    }

    /* Complex getters (still non-mutating) */

    /**
     * @param $link
     *
     * @return string ('creator'|'unlocked'|'not-locked'|'locked'|'expired'|'no-cookies'|'no-user')
     */
    private static function get_lock_status($link)
    {
        if (array_key_exists('cookiesNotAllowed', $_GET)) {
            return 'no-cookies';
        }
        if (!self::is_user_logged_in()) {
            return 'no-user';
        }

        $lockedPostObject = self::get_locked_post_object($link);

        if (!$lockedPostObject) {
            return 'not-locked';
        }

        switch ($lockedPostObject ['linkState']) {
            case 'READER_LINK_STATE_CREATOR':
                return 'creator';
            case 'READER_LINK_STATE_PURCHASED':
                return 'unlocked';
            case 'READER_LINK_STATE_EXPIRED':
                return 'expired';
            default:
                if (self::user_can_edit_posts()) {
                    return 'creator';
                }

                return 'locked';
        }
    }

    /**
     * Format url (remove http etc) to match Lowest Common Denominator.
     *
     * @param string (url)
     *
     * @return string (url)
     */
    public static function get_common_url($url)
    {
        $disallowed = ['http://', 'https://', 'www.'];
        foreach ($disallowed as $d) {
            $start_regex = '|^'.$d.'|';
            $url = preg_replace($start_regex, '', $url);
        }
        // Remove query fragment
        $url = preg_replace('|\?.*?$|', '', $url);
        // Remove trailing slash
        $url = preg_replace('|/$|', '', $url);

        return esc_url_raw("https://$url", ['https']);
    }

    /**
     * Return diff expire date and now.
     *
     * @param $expireDate
     *
     * @return string diff number plus unit (days/hours)
     */
    public static function get_time_diff($expireDate)
    {
        $datetime1 = date_create($expireDate);
        $datetime2 = date_create('now');
        $interval = date_diff($datetime1, $datetime2);

        $hours = $interval->format('%h');
        $days = $interval->format('%a');

        if ($days > 0) {
            $number = (int) round(1 + (int) $days);

            return sprintf(
                esc_html(_n('%s day', '%s days', $number, 'exodox')),
                $number
            );
        }

        $number = (int) round($hours);

        return sprintf(
            esc_html(_n('%s hour', '%s hours', $number, 'exodox')),
            $number
        );
    }

    /**
     * Converts ms to hours or days.
     *
     * @param $millis
     *
     * @return string number plus unit (days/hours)
     */
    public static function get_days_hours($millis)
    {
        $days = ($millis / 86400000);
        if ($days < 2) {
            $number = (int) round($days * 24);

            return sprintf(
                esc_html(_n('%s hour', '%s hours', $number, 'exodox')),
                $number
            );
        }

        $number = (int) round($days);

        return sprintf(
            esc_html(_n('%s day', '%s days', $number, 'exodox')),
            $number
        );
    }

    /**
     * Returns locked post assoc array based on url.
     *
     * @param string (link to check)
     *
     * @return array (locked post data)
     */
    public static function get_locked_post_object($link)
    {
        $cleanLink = self::get_common_url($link);
        if (array_key_exists($cleanLink, self::$exodoxLockedPostsObjects)) {
            return self::$exodoxLockedPostsObjects[$cleanLink];
        }

        return null;
    }

    /**
     * Get post first image.
     *
     * @param object (post)
     *
     * @return array img url and alt-text
     */
    public static function get_first_image($post) // TODO: This function only works if is_single() is true - check!
    {
        $output = preg_match('/<img.*?src\s*=\s*"(.*?)".*?alt\s*=\s*"(.*?)".*?>/i', $post->post_content, $matches);
        if ($output) {
            return [
                'src' => esc_attr($matches[1]),
                'alt' => esc_attr($matches[2]),
            ];
        }

        return null;
    }

    /* Content generators (decidedly more complex getters, should still be non-mutating) */

    /**
     * Extracts and returns image data for locked post.
     *
     * @param $post
     * @param $lockedPostObject
     *
     * @return array
     */
    private static function get_image_data($post, $link)
    {
        $lockedPostObject = self::get_locked_post_object($link);
        $lockStatus = self::get_lock_status($link);
        $postID = $post->ID;
        $featuredImageId = get_post_thumbnail_id($postID);
        if ($featuredImageId) {
            $featuredImageData = get_post($featuredImageId);
            $featuredImageAlt = $featuredImageData->post_excerpt ? $featuredImageData->post_excerpt : '';
            $featuredImage = wp_get_attachment_image_src($featuredImageId, 'single-post-thumbnail');
        }

        $articleImage = $lockedPostObject ['image'];
        $articleImageAlt = '';
        $hideDefaultMediaCss = '';
        // TODO: We don't get any alt-text from Exodox

        if ('' === $articleImage) {
            ['src' => $articleImage, 'alt' => $articleImageAlt] = self::get_first_image($post);
        }

        if ($featuredImageId && 'unlocked' !== $lockStatus && 'creator' !== $lockStatus) {
            $articleImage = $featuredImage[0];
            $articleImageAlt = $featuredImageAlt;
            $hideDefaultMediaCss = "#post-$postID > .post-media { display:none; }";
        }

        return [
            'articleImage' => $articleImage,
            'articleImageAlt' => $articleImageAlt,
            'mediaCss' => $hideDefaultMediaCss,
        ];
    }

    /**
     * Generates content html for locked post.
     * TODO: This is probably nicer to have in a template.
     *
     * @param array $params (texts, links, and classes for the generated html)
     *
     * @return string
     */
    private static function generate_single_locked_item_html($params)
    {
        $top_message = wp_kses_post($params ['top_message']);
        $action = wp_kses_post($params ['action']);
        $bottom_message = isset($params ['bottom_message']) ? wp_kses_post($params ['bottom_message']) : null;
        $shortcuts = wp_kses_post('<li>'.join('</li><li>', $params ['shortcuts']));

        if (array_key_exists('priceText', $params)) {
            $priceText = esc_html($params ['priceText']);
            $validTime = esc_html($params ['validTime']);

            $priceInfo = <<< HTML
                <div class='exodox-top-info'>
                    <p class='exodox-top-info-price-wrapper'>
                        <span class='exodox-top-info-price'>$priceText</span>
                        <span class='exodox-top-info-terms'>$validTime</span>
                    </p>
                </div>
HTML;
        } else {
            $priceInfo = '';
        }

        if (array_key_exists('mediaCss', $params)) {
            wp_add_inline_style('exodoxCss', $params ['mediaCss']);
        }
        $permaLink = esc_url($params ['permalink']);
        $slug = esc_attr($params ['slug']);
        $title = esc_html($params ['title']);
        $imgSrc = esc_url($params ['imageSrc']);
        $imgAlt = esc_attr($params ['imageAlt']);
        $preamble = esc_html($params ['preamble']);
        $abuse_action = esc_url(sprintf(
            __('%1$sReport abuse link%2$sI would like to report the Exodox link %3$s with the description: %4$s', 'exodox'),
            'mailto:hello@exodox.link?subject=',
            '&body=',
            $permaLink,
            $slug
        ));
        $abuse_text = esc_html__('Report abuse', 'exodox');

        $content = <<< HTML
            <div class="exodox-wrapper">
                <div class="exodox-wrapper-top">$top_message</div>
                <div class="exodox-wrapper-border">
                    $priceInfo
                    <h2 class="entry-title">
                        <a class="exodox-no-underline" href="$permaLink">$title</a>
                    </h2>
                    <img class="exodox-teaser-image" src="$imgSrc" alt="$imgAlt"><br>
                    <p>$preamble</p>
                    $action
                </div>
                <div class="exodox-wrapper-bottom">$bottom_message</div>
                <ul class="exodox-wrapper-shortcuts">$shortcuts</ul>
                <div class="exodox-wrapper-bottom"><a href="$abuse_action">$abuse_text</a></div>
                
            </div>
HTML;
        // Replacing new-lines in WP is important, leaving them in will turn them into p-tags
        return str_replace(["\n", "\r"], ' ', $content);
    }

    /**
     * Set the correct content depending on lock status for a single article.
     *
     * @param $post
     *
     * @return object (post)
     */
    private static function lock_single_post($post)
    {
        $link = self::get_common_url(get_permalink($post->ID));
        $status = self::get_lock_status($link);
        $lockedPostObject = self::get_locked_post_object($link);
        $price = $lockedPostObject ['price'] * 0.01;
        $currency = $lockedPostObject ['currency'];
        $permalink = $lockedPostObject ['queryLink'];
        $slug = $lockedPostObject ['slug'];
        $lockHeading = __('Unlock', 'exodox');

        [
            'articleImage' => $articleImage,
            'articleImageAlt' => $articleImageAlt,
            'mediaCss' => $hideDefaultMediaCss,
        ] = self::get_image_data($post, $link);

        $preamble = $lockedPostObject ['preamble'];
        $validTimeLength = self::get_days_hours($lockedPostObject ['validTimeLength']);
        $userId = self::get_user_id();

        $link_unlock = $userId
            ? esc_url_raw(self::EXODOX_TEASER_PAGE.$slug."?autounlock=$userId", ['https'])
            : esc_url_raw(self::EXODOX_TEASER_PAGE.$slug.'?loginUrl='.urlencode(site_url().'/'.self::EXODOX_CHECK_LOGIN_RETURN_PAGE.'&targetUrl='.self::get_current_url()), ['https']);

        $action = sprintf(
            $userId
                ? __('%1$sUnlock Link%2$s%3$s', 'exodox')
                : __('%1$sUnlock Link%2$sLogin to Exodox%3$s', 'exodox'),
            "<a href=\"$link_unlock\" class=\"exodox-cta\"><span class=\"exodox-cta-heading\" data-testid=\"$status\">",
            '</span>',
            '</a>'
        );

        $bottom_message = $userId
            ? sprintf(
                __('<em>%1$s %2$s</em> will be added to your monthly billing', 'exodox'),
                $price,
                $currency
            )
            : null;

        $shortcut_create = sprintf(
            __('New to Exodox?%1$sCreate Account%2$s', 'exodox'),
            '<br /><a href="'.esc_url_raw(self::EXODOX_LOGIN_PAGE."?linkPreviewSlug={$slug}", ['https']).'">',
            '</a>'
        );

        $shortcut_dashboard = sprintf(
            __('%1$sGo to my dashboard%2$s', 'exodox'),
            '<a href="'.esc_url_raw(self::EXODOX_TEASER_PAGE, ['https']).'">',
            '</a>'
        );

        $shortcut_logout = sprintf(
            __('%1$sLogout%2$s', 'exodox'),
            '<a href="#">',
            '</a>'
        );

        switch ($status) {
            case 'not-locked':
                return $post;
            case 'creator':
                $content = '<p class="exodox-pre-wrapper-info" data-testid="'.$status.'">'.esc_html__('The article is locked with Exodox; but you, as administrator, can still read it.', 'exodox').'</p>'.$post->post_content;
                break;
            case 'unlocked':
                $validTime = self::get_time_diff(urldecode($lockedPostObject ['expiresAt']));
                $content = '<p class="exodox-pre-wrapper-info" data-testid="'.$status.'">'.sprintf(esc_html__('The article is unlocked. Read the entire article for %s more', 'exodox'), $validTime).'</p>'.$post->post_content;
                break;
            case 'no-cookies':
                $content = self::generate_single_locked_item_html([
                    'top_message' => __('You must <em>allow cookies</em> to be able to unlock this article', 'exodox'),
                    'action' => $action,
                    'shortcuts' => [$shortcut_create],
                    'mediaCss' => $hideDefaultMediaCss,
                    'priceText' => sprintf(__('%1$s %2$s', 'exodox'), $price, $currency),
                    'validTime' => sprintf(__('%s access', 'exodox'), $validTimeLength),
                    'permalink' => $permalink,
                    'slug' => $slug,
                    'title' => $post->post_title,
                    'imageSrc' => $articleImage,
                    'imageAlt' => $articleImageAlt,
                    'preamble' => $preamble,
                ]);
                break;
            case 'no-user':
                $content = self::generate_single_locked_item_html([
                    'top_message' => sprintf(
                        __('%1$sYou need an Exodox account to be able to purchase access to this link. %2$sWhat is Exodox?%3$s%4$sLogin to unlock%5$s', 'exodox'),
                        '<p>',
                        '<a href="'.esc_url_raw('https://exodox.link/consumer', ['https']).'">',
                        '</a></p>',
                        "<ul class=\"exodox-wrapper-shortcuts\"><li><a href=\"$link_unlock\">",
                        '</a></li></ul>'
                    ),
                    'action' => $action,
                    'shortcuts' => [$shortcut_create],
                    'mediaCss' => $hideDefaultMediaCss,
                    'priceText' => sprintf(__('%1$s %2$s', 'exodox'), $price, $currency),
                    'validTime' => sprintf(__('%s access', 'exodox'), $validTimeLength),
                    'permalink' => $permalink,
                    'slug' => $slug,
                    'title' => $post->post_title,
                    'imageSrc' => $articleImage,
                    'imageAlt' => $articleImageAlt,
                    'preamble' => $preamble,
                ]);
                break;
            case 'expired':
                $content = self::generate_single_locked_item_html([
                    'top_message' => sprintf(
                        __('Unfortunately, your previous purchase has expired. %1$sBuy again%2$s', 'exodox'),
                        "<a href=\"$link_unlock\">",
                        '</a>'
                    ),
                    'action' => $action,
                    'bottom_message' => $bottom_message,
                    'shortcuts' => [$shortcut_dashboard],
                    'mediaCss' => $hideDefaultMediaCss,
                    'priceText' => sprintf(__('%1$s %2$s', 'exodox'), $price, $currency),
                    'validTime' => sprintf(__('%s access', 'exodox'), $validTimeLength),
                    'permalink' => $permalink,
                    'slug' => $slug,
                    'title' => $post->post_title,
                    'imageSrc' => $articleImage,
                    'imageAlt' => $articleImageAlt,
                    'preamble' => $preamble,
                ]);
                break;
            default:
                $content = self::generate_single_locked_item_html([
                    'top_message' => sprintf(
                        __('Logged in.', 'exodox'), // 'Logged in as %1$s. %2$s(Log out)%3$s'
                        '',
                        '<a href="#">',
                        '</a>'
                    ),
                    'action' => $action,
                    'bottom_message' => $bottom_message,
                    'shortcuts' => [$shortcut_dashboard],
                    'mediaCss' => $hideDefaultMediaCss,
                    'priceText' => sprintf(__('%1$s %2$s', 'exodox'), $price, $currency),
                    'validTime' => sprintf(__('%s access', 'exodox'), $validTimeLength),
                    'permalink' => $permalink,
                    'slug' => $slug,
                    'title' => $post->post_title,
                    'imageSrc' => $articleImage,
                    'imageAlt' => $articleImageAlt,
                    'preamble' => $preamble,
                ]);
        }

        $post->is_exodox_locked = isset($contentLockLink);
        $post->post_content = $content;

        return $post;
    }

    /* Render functions */

    /**
     * Render admin page.
     */
    public static function render_page()
    {
        if (!current_user_can('manage_options')) {
            return;
        }

        settings_errors(self::EXODOX_ADMIN_PAGE_SLUG); ?>
        <div class='wrap'>
            <h1><?php echo esc_html(get_admin_page_title()); ?></h1>
            <form action='options.php' method='post'>
                <?php
                do_settings_sections(self::EXODOX_ADMIN_PAGE_SLUG);
        settings_fields(self::EXODOX_ADMIN_PAGE_SLUG);
        submit_button(esc_html__('Update', 'exodox')); ?>
            </form>
        </div>
        <?php
    }

    /**
     * Output input field.
     *
     * @param array $args Input field properties
     */
    public static function input_field(array $args)
    {
        $name = array_key_exists('name', $args) ? $args ['name'] : '';

        printf(
            '<input type="%s" class="%s" name="%s" value="%s" placeholder="%s" />%s',
            esc_attr(array_key_exists('type', $args) ? $args ['type'] : 'text'),
            esc_attr(array_key_exists('class', $args) ? $args ['class'] : ''),
            esc_attr($name),
            esc_attr(get_option($name, '')),
            esc_attr(array_key_exists('placeholder', $args) ? $args ['placeholder'] : ''),
            array_key_exists('help_text', $args) ? sprintf('<p class="description">%s</p>', esc_html($args ['help_text'])) : ''
        );
    }

    /**
     * Output checkbox to toggle display of elements on listing pages.
     *
     * @since 0.7.8
     *
     * @param array $args Checkbox properties
     *
     * @return null
     */
    public static function exodox_disable_listing_checkbox(array $args)
    {
        printf(
            '<input type="checkbox" class="%s" name="exodox_disable_listing" value="yes" %s />%s',
            esc_attr(array_key_exists('class', $args) ? $args ['class'] : ''),
            esc_attr('yes' === get_option('exodox_disable_listing') ? 'checked=checked' : ''),
            array_key_exists('help_text', $args) ? sprintf('<p class="description">%s</p>', esc_html($args ['help_text'])) : ''
        );
    }

    /* Functions calling wp api to add style/script files or tags */

    /**
     * Enqueue css for single locked post.
     *
     * @param $post
     */
    private static function enqueue_lock_single_css($post)
    {
        $postID = $post->ID;
        $css = <<<CSS
            #post-$postID img, .post-$postID img
            {
                display: none;
            }
            #post-$postID .exodox-wrapper-border img, .post-$postID .exodox-wrapper-border img
            {
                display: block;
                width: 100%;
            }
            #post-$postID .entry-title, .post-$postID .entry-title, #post-$postID .post-title, .post-$postID .post-title {
                display: none;
            }
            #post-$postID .exodox-wrapper-border .entry-title, .post-$postID .exodox-wrapper-border .entry-title {
                display: block;
            }
CSS;

        wp_add_inline_style('exodoxCss', $css);
    }

    /* Functions calling external APIs */

    /**
     * Retrieve list of locked posts given link input.
     *
     * @global string wp_version WordPress version
     *
     * @param $links
     *
     * @return array
     */
    private static function retrieve_locked_posts($links)
    {
        $empty_result = ['links' => []];
        if (!self::get_publisher_id()) {
            return $empty_result;
        }
        $content = ['urls' => $links];
        if (self::is_user_logged_in()) {
            $content ['readerId'] = self::get_user_id();
        }

        global $wp_version;
        $options = [
            'method' => 'POST',
            'headers' => [
                'Content-Type' => 'application/json',
                'Authorization' => 'Creator '.self::get_publisher_id(),
                'Accept' => 'application/json',
                'User-Agent' => 'ExodoxClient/1.0 (WordPress; WordPress/'.$wp_version.' PHP/'.phpversion().') ExodoxWordPressPlugin/'.self::EXODOX_PLUGIN_VERSION,
            ],
            'timeout' => self::EXODOX_API_TIMEOUT,
            'body' => json_encode($content, JSON_UNESCAPED_SLASHES | JSON_UNESCAPED_UNICODE),
            'sslverify' => EXODOX_DEV ? false : true,
        ];

        $endpoint = self::is_user_logged_in()
            ? $endpoint = self::EXODOX_CHECK_LOCKED_POSTS_WITH_UUID_ENDPOINT
            : self::EXODOX_CHECK_LOCKED_POSTS_ENDPOINT;

        // TODO: At this point, it might be useful to check a cache

        $response = wp_remote_get($endpoint, $options);
        $status_code = wp_remote_retrieve_response_code($response);

        if (!is_wp_error($response) && 200 === $status_code) {
            $body = wp_remote_retrieve_body($response);

            $parsedResult = json_decode($body, true);

            if (0 !== json_last_error()) {
                self::log_warning("ERROR:malformed json return from OpenAPI\n".self::format_data_raw(json_last_error_msg()));

                return $empty_result;
            }
            if (!is_array($parsedResult)) {
                self::log_warning("ERROR: malformed data return from OpenAPI\n".self::format_data_raw($parsedResult));

                return $empty_result;
            }
            if (!array_key_exists('links', $parsedResult) || !is_array($parsedResult ['links'])) {
                self::log_warning("ERROR: missing links in return value from OpenAPI\n".self::format_data_raw($parsedResult));

                return $empty_result;
            }

            return $parsedResult;
        }

        self::log_warning("ERROR: OpenAPI responded with an error (or timed out)\n");

        return $empty_result;
    }

    /* Parameter checks (should not mutate the state, but may log or kill the session) */

    /**
     * Checks for mandatory parameters to the exodox login endpoint.
     *
     * @uses $_GET ['targetUrl']
     */
    private static function check_login_return_parameters()
    {
        if (!array_key_exists('targetUrl', $_GET)) {
            wp_die('Missing parameter targetUrl', null, ['response' => 401]);
        }
    }

    /* Redirect functions */

    /**
     * Makes safe (302) redirect to post.
     *
     * @param $postLink
     */
    public static function redirect_to_post($postLink)
    {
        wp_safe_redirect($postLink, 302);
        exit;
    }

    /**
     * Makes safe (302) redirect to post indicating cookies are not allowed.
     *
     * @uses $_GET ['targetUrl']
     */
    private static function redirect_to_target_url_no_cookies()
    {
        $cookieParameter = ['cookiesNotAllowed' => 1];
        $targetUrl = esc_url_raw($_GET ['targetUrl'], ['https']);
        $targetUrl = add_query_arg($cookieParameter, urldecode($targetUrl));
        wp_safe_redirect($targetUrl);
        exit;
    }

    /**
     * Makes permanent (301) redirect to same url (incl query params) on https.
     *
     * @uses $_SERVER['REQUEST_URI']
     */
    private static function redirect_to_https()
    {
        $site_url = get_site_url(null, '', 'https');
        $url = $site_url.$_SERVER['REQUEST_URI'];
        wp_safe_redirect($url, 301);
        exit;
    }

    /**
     * Prepare URL and redirect to targetUrl.
     *
     * @uses $_GET ['targetUrl']
     */
    public static function redirect_to_target_url()
    {
        $targetUrl = esc_url_raw($_GET ['targetUrl'], ['https']);
        self::redirect_to_post(urldecode($targetUrl));
    }

    /**
     * Makes safe (302) redirect to login check page.
     */
    public static function redirect_to_login_check()
    {
        setcookie(self::EXODOX_COOKIE_NAME, 'none', time() + (86400 * 30), '/');
        wp_safe_redirect(self::EXODOX_LOGIN_CHECK_PAGE.'?loginUrl='.urlencode(site_url().'/'.self::EXODOX_CHECK_LOGIN_RETURN_PAGE).'&targetUrl='.self::get_current_url().'&cookieSet=1', 302);
        exit;
    }

    /**********************************************/
    /* Debug functions (should be private static) */
    /**********************************************/

    /**
     * Return html formated data for debugging purposes.
     *
     * @param $data
     *
     * @return string
     */
    private static function format_data($data)
    {
        return '<pre style="border: 2px solid red;">'.preg_replace('/</', '&lt;', print_r($data, true)).'</pre>';
    }

    /**
     * Return formated data for debugging purposes.
     *
     * @param $data
     *
     * @return string
     */
    private static function format_data_raw($data)
    {
        return print_r($data, true);
    }

    /**
     * Log error.
     *
     * @param $message
     */
    private static function log_error($message)
    {
        error_log($message, 0);
    }

    /**
     * Log warning, if in debug mode.
     *
     * @param $message
     */
    private static function log_warning($message)
    {
        if (!WP_DEBUG) {
            return;
        }
        trigger_error($message, E_USER_WARNING);
    }
}

ExodoxLogin::initialize();
